from contextlib import redirect_stdout
from io import StringIO

import pytest
import tiktoken
from fast_bpe_rs import BPE
from tiktoken._educational import SimpleBytePairEncoding


def _train_tiktoken_encoding(
    training_doc: str,
    split_pattern: str,
    vocab_size: int,
    special_tokens: dict[str, int] | None = None,
) -> tiktoken.Encoding:
    with redirect_stdout(StringIO()):
        educational_bpe = SimpleBytePairEncoding.train(
            training_doc,
            vocab_size=vocab_size,
            pat_str=split_pattern,
        )

    return tiktoken.Encoding(
        name="fast-bpe-rs-test",
        pat_str=split_pattern,
        mergeable_ranks=educational_bpe.mergeable_ranks,
        special_tokens=special_tokens or {},
    )


def _assert_fast_bpe_matches_tiktoken(
    training_doc: str,
    sample: str,
    split_pattern: str,
    vocab_size: int,
    special_tokens: dict[str, int] | None = None,
) -> None:
    fast_bpe = BPE(split_pattern, special_tokens)
    fast_bpe.train(vocab_size, [training_doc])

    tiktoken_bpe = _train_tiktoken_encoding(
        training_doc,
        split_pattern,
        vocab_size,
        special_tokens,
    )

    encode_kwargs = {"allowed_special": "all"} if special_tokens else {}
    fast_encoded = fast_bpe.encode(sample)
    tiktoken_encoded = tiktoken_bpe.encode(sample, **encode_kwargs)

    assert fast_encoded == tiktoken_encoded
    assert fast_bpe.decode(fast_encoded) == tiktoken_bpe.decode_bytes(tiktoken_encoded)
    assert fast_bpe.decode_to_string(fast_encoded) == tiktoken_bpe.decode(
        tiktoken_encoded
    )


@pytest.mark.parametrize(
    ("training_doc", "sample"),
    [
        ("abababa", "abababa"),
        ("abababa", "ababa"),
        ("banana banana", "banana banana"),
        ("banana banana", "banana"),
        ("go go stop", "go go stop"),
        ("go go stop", "go stop"),
        ("mississippi river", "mississippi river"),
        ("mississippi river", "river"),
    ],
)
def test_fast_bpe_rs_matches_tiktoken_for_encoding_and_decoding(
    training_doc: str,
    sample: str,
) -> None:
    _assert_fast_bpe_matches_tiktoken(
        training_doc=training_doc,
        sample=sample,
        split_pattern=r"(?s).+",
        vocab_size=257,
    )


@pytest.mark.parametrize(
    ("training_doc", "sample", "special_tokens"),
    [
        (
            "banana banana",
            "banana<pad><eos> banana",
            {"<pad>": 900, "<eos>": 901},
        ),
        (
            "go go stop",
            "<pad>go<end?>stop<pad>",
            {"<pad>": 800, "<end?>": 801},
        ),
    ],
)
def test_fast_bpe_rs_matches_tiktoken_with_special_tokens(
    training_doc: str,
    sample: str,
    special_tokens: dict[str, int],
) -> None:
    _assert_fast_bpe_matches_tiktoken(
        training_doc=training_doc,
        sample=sample,
        split_pattern=r"(?s).+",
        vocab_size=257,
        special_tokens=special_tokens,
    )



"""Train fast_bpe_rs on WikiText with Hugging Face datasets."""

from __future__ import annotations

import argparse
import time

from datasets import load_dataset
from fast_bpe_rs import BPE

DATASET_REPO = "Salesforce/wikitext"
DEFAULT_DATASET_CONFIG = "wikitext-103-raw-v1"
SUPPORTED_DATASET_CONFIGS = ("wikitext-103-raw-v1", "wikitext-2-raw-v1")
DATASET_SPLIT = "train"
DEFAULT_TARGET_VOCAB_SIZE = 1 << 15
DEFAULT_RUNS = 3

REGEX = (
    r"'(?i:[sdmt]|ll|ve|re)|[^\r\n\p{L}\p{N}]?+\p{L}++|\p{N}{1,3}+"
    r"| ?[^\s\p{L}\p{N}]++[\r\n]*+|\s++$|\s*[\r\n]|\s+(?!\S)|\s"
)


def positive_int(value: str) -> int:
    parsed = int(value)
    if parsed <= 0:
        raise argparse.ArgumentTypeError("value must be a positive integer")
    return parsed


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Train fast_bpe_rs on the WikiText train split and print timing."
    )
    parser.add_argument(
        "--dataset-config",
        choices=SUPPORTED_DATASET_CONFIGS,
        default=DEFAULT_DATASET_CONFIG,
        help=(
            f"WikiText dataset config to train on (default: {DEFAULT_DATASET_CONFIG})."
        ),
    )
    parser.add_argument(
        "--target-vocab-size",
        type=positive_int,
        default=DEFAULT_TARGET_VOCAB_SIZE,
        help=(
            "Target vocabulary size for BPE training "
            f"(default: {DEFAULT_TARGET_VOCAB_SIZE})."
        ),
    )
    parser.add_argument(
        "--runs",
        type=positive_int,
        default=DEFAULT_RUNS,
        help=f"Number of training runs to execute (default: {DEFAULT_RUNS}).",
    )
    return parser.parse_args()


def load_wikitext_train_docs(dataset_config: str) -> tuple[list[str], int]:
    dataset = load_dataset(DATASET_REPO, dataset_config, split=DATASET_SPLIT)
    total_rows = len(dataset)
    docs = [text.strip() for text in dataset["text"] if text and text.strip()]
    return docs, total_rows


def main() -> None:
    args = parse_args()
    docs, total_rows = load_wikitext_train_docs(args.dataset_config)
    print(
        "RUN_CONTEXT "
        f"dataset_repo={DATASET_REPO} "
        f"dataset_config={args.dataset_config} "
        f"dataset_split={DATASET_SPLIT} "
        f"split_rows={total_rows} "
        f"docs_loaded={len(docs)} "
        f"rows_filtered_out={total_rows - len(docs)} "
        f"target_vocab_size={args.target_vocab_size} "
        f"runs={args.runs}"
    )

    for run in range(1, args.runs + 1):
        bpe = BPE(REGEX)
        start_ns = time.perf_counter_ns()
        bpe.train(args.target_vocab_size, docs)
        elapsed_ms = (time.perf_counter_ns() - start_ns) // 1_000_000
        print(f"TRAIN_RUN run={run} elapsed_ms={elapsed_ms}")

    print("TRAINING_COMPLETE finished=true")


if __name__ == "__main__":
    main()
